# Written by: Readman Chiu
# Edited by: Lucas Swanson

import os, re
from parsers import gene_annotation
from utils.general import IsInteger

class Transcript:

    name = None
    chrom = None
    strand = None
    txStart = txEnd = None
    cdsStart = cdsEnd = None
    exonCount = None
    exons = None
    gene = None
    alias = None
    model = None
    weight = None
    length = 0

    def __init__(self, name):
        self.name = name

    def full_name(self):
        alias = "NA"
        if self.alias:
            alias = self.alias
        
        return "%s(%s)" % (self.name, alias)

    def details(self):
        return "%s %s %s %s-%s %s" % (self.full_name(), self.chrom, self.strand, self.txStart, self.txEnd, self.exonCount)

    def exon_string(self):
        return ",".join(["%i-%i" % (exon[0], exon[1]) for exon in self.exons])

    def cds_length(self):
        if self.cdsStart and self.cdsEnd and IsInteger(self.cdsStart) and IsInteger(self.cdsEnd):
            return int(self.cdsEnd) - int(self.cdsStart)

        return 0
    
    def coding_type(self):
        if self.cdsStart == None or self.cdsEnd == None:
            return 'NA'
        elif IsInteger(self.cdsStart) and IsInteger(self.cdsEnd) and self.cdsStart != self.cdsEnd:
            return 'CODING'
        else:
            return 'NONCODING'

    @classmethod
    def find_by_name(cls, model, annot_file, txt_name):
        txts = []

        if os.path.exists(annot_file):
            for line in open(annot_file, 'r'):
                if txt_name in line:
                    txt = {
                        'e': gene_annotation.ensembl.parse_line,
                        'r': gene_annotation.refGene.parse_line,
                        'k': gene_annotation.knownGene.parse_line,
                        'a': gene_annotation.aceview.parse_line,
                        'x': gene_annotation.ensg.parse_line
                        }[model](line)

                    if txt.name == txt_name or txt.alias == txt_name:
                        txts.append(txt)

        return txts
